from __future__ import annotations

import sys
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Literal, TypeVar

import openapi_fal_rest.api.files.file_exists as file_exists_api
import openapi_fal_rest.models.file_spec as file_spec_model
from fal.rest_client import REST_CLIENT
from fal.toolkit import mainify

if sys.version_info >= (3, 11):
    from typing import Concatenate
else:
    from typing_extensions import Concatenate

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

import fal.flags as flags
from fal.api import FalServerlessError, InternalFalServerlessError, function

ArgsT = ParamSpec("ArgsT")
ReturnT = TypeVar("ReturnT")


def file_exists(
    file_path: Path | str, calculate_checksum: bool
) -> file_spec_model.FileSpec | None:
    file_str = str(file_path)
    response = file_exists_api.sync_detailed(
        file_str,
        client=REST_CLIENT,
        calculate_checksum=calculate_checksum,
    )

    if response.status_code == 200:
        return response.parsed  # type: ignore
    elif response.status_code == 404:
        return None
    else:
        raise InternalFalServerlessError(
            f"Failed to check if file exists: {response.status_code} {response.parsed}"
        )


@mainify
@dataclass
class FileSpec:
    path: Path
    size: int
    is_file: bool
    checksum_sha256: str | None = None
    checksum_md5: str | None = None

    @classmethod
    def from_model(cls, file_spec: file_spec_model.FileSpec) -> FileSpec:
        return cls(
            path=Path(file_spec.path),
            size=file_spec.size,
            is_file=file_spec.is_file,
            checksum_sha256=file_spec.checksum_sha256
            if file_spec.checksum_sha256
            else None,
            checksum_md5=file_spec.checksum_md5 if file_spec.checksum_md5 else None,
        )


def setup(
    file_path: Path | str,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
    **isolated_config: Any,
):
    file_path = Path(file_path)
    check_path = file_path.relative_to("/data")

    def wrapper(
        func: Callable[Concatenate[ArgsT], ReturnT]
    ) -> Callable[Concatenate[ArgsT], Path]:
        @wraps(func)
        def internal_wrapper(
            *args: ArgsT.args,
            **kwargs: ArgsT.kwargs,
        ) -> Path:
            checksum = bool(checksum_sha256 or checksum_md5)

            file = file_exists(check_path, calculate_checksum=checksum)

            if not file or force or flags.FORCE_SETUP:
                config = {
                    "machine_type": "S",
                    **isolated_config,
                }
                function(**config)(func)(*args, **kwargs)  # type: ignore

                file = file_exists(check_path, calculate_checksum=checksum)

            if not file:
                raise FalServerlessError(
                    f"Setup function {func.__name__} did not create expected path."
                )

            if checksum:
                if not file.is_file:
                    raise FalServerlessError(
                        f"Setup function {func.__name__} created a directory instead of a file."
                    )

                if checksum_sha256 and file.checksum_sha256 != checksum_sha256:
                    raise FalServerlessError(
                        f"Setup function {func.__name__} created file with unexpected SHA256 checksum.\n"
                        f"Expected {checksum_sha256} but got {file.checksum_sha256}"
                    )

                if checksum_md5 and file.checksum_md5 != checksum_md5:
                    raise FalServerlessError(
                        f"Setup function {func.__name__} created file with unexpected MD5 checksum.\n"
                        f"Expected {checksum_md5} but got {file.checksum_md5}"
                    )

            return Path(file.path)

        return internal_wrapper

    return wrapper


DownloadType = Literal["python", "wget", "curl"]


def download_file(
    url: str,
    check_location: str | Path,
    *,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
    tool: DownloadType = "python",
):
    check_path = Path(check_location)

    @setup(
        check_path,
        checksum_sha256,
        checksum_md5,
        force=force,
        # isolated configs
        requirements=["urllib3"],
    )
    def download(force: bool):
        import os
        from shutil import copyfileobj
        from urllib.request import Request, urlopen

        print(f"Downloading {url} to {check_path}")

        # Make sure the directory exists
        check_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            # TODO: how can we randomize the user agent to avoid being blocked?
            user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:21.0) Gecko/20100101 Firefox/21.0"

            req = Request(url, headers={"User-Agent": user_agent})

            if check_path.exists() and not force:
                print(f"File {check_path} already exists, skipping download")
                return

            if tool == "python":
                with urlopen(req) as response, check_path.open("wb") as f:
                    copyfileobj(response, f)
            elif tool == "curl":
                command = f"curl {req.full_url} -o {check_path}"

                print(command)
                res = os.system(command)

                if res != 0:
                    raise Exception(f"curl failed with exit code {res}")
            elif tool == "wget":
                command = f"wget {req.full_url} -O {check_path}"

                print(command)
                res = os.system(command)

                if res != 0:
                    raise Exception(f"wget failed with exit code {res}")
            else:
                raise Exception(f"Unknown download tool {tool}")

        except Exception as e:
            msg = f"Failed to download {url} to {check_path}\n{e}"
            print(msg)

            os.system(f"rm -rf {check_path}")

            # raise exception in generally-available class
            raise FileNotFoundError(msg) from None

    return download(force)


def download_weights(
    url: str,
    *,
    checksum_sha256: str | None = None,
    checksum_md5: str | None = None,
    force: bool = False,
    tool: DownloadType = "python",
):
    import hashlib

    url_id = hashlib.sha256(url.encode("utf-8")).hexdigest()
    # This is not a protected path, so the user may change stuff internally
    url_path = Path(f"/data/.fal/downloads/{url_id}")

    return download_file(
        url,
        url_path,
        checksum_sha256=checksum_sha256,
        checksum_md5=checksum_md5,
        force=force,
        tool=tool,
    )



import datetime
from typing import Any, Dict, List, Type, TypeVar

import attr
from dateutil.parser import isoparse

from ..models.key_scope import KeyScope

T = TypeVar("T", bound="UserKeyInfo")


@attr.s(auto_attribs=True)
class UserKeyInfo:
    """
    Attributes:
        key_id (str):
        created_at (datetime.datetime):
        scope (KeyScope): An enumeration.
        alias (str):
    """

    key_id: str
    created_at: datetime.datetime
    scope: KeyScope
    alias: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        key_id = self.key_id
        created_at = self.created_at.isoformat()

        scope = self.scope.value

        alias = self.alias

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "key_id": key_id,
                "created_at": created_at,
                "scope": scope,
                "alias": alias,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        key_id = d.pop("key_id")

        created_at = isoparse(d.pop("created_at"))

        scope = KeyScope(d.pop("scope"))

        alias = d.pop("alias")

        user_key_info = cls(
            key_id=key_id,
            created_at=created_at,
            scope=scope,
            alias=alias,
        )

        user_key_info.additional_properties = d
        return user_key_info

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..models.invoice_item import InvoiceItem


T = TypeVar("T", bound="Invoice")


@attr.s(auto_attribs=True)
class Invoice:
    """
    Attributes:
        amount_due (int):
        period_start (int):
        period_end (int):
        subtotal (int):
        total (int):
        items (List['InvoiceItem']):
        url (str):
        starting_balance (int):
        status (str):
    """

    amount_due: int
    period_start: int
    period_end: int
    subtotal: int
    total: int
    items: List["InvoiceItem"]
    url: str
    starting_balance: int
    status: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        amount_due = self.amount_due
        period_start = self.period_start
        period_end = self.period_end
        subtotal = self.subtotal
        total = self.total
        items = []
        for items_item_data in self.items:
            items_item = items_item_data.to_dict()

            items.append(items_item)

        url = self.url
        starting_balance = self.starting_balance
        status = self.status

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "amount_due": amount_due,
                "period_start": period_start,
                "period_end": period_end,
                "subtotal": subtotal,
                "total": total,
                "items": items,
                "url": url,
                "starting_balance": starting_balance,
                "status": status,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.invoice_item import InvoiceItem

        d = src_dict.copy()
        amount_due = d.pop("amount_due")

        period_start = d.pop("period_start")

        period_end = d.pop("period_end")

        subtotal = d.pop("subtotal")

        total = d.pop("total")

        items = []
        _items = d.pop("items")
        for items_item_data in _items:
            items_item = InvoiceItem.from_dict(items_item_data)

            items.append(items_item)

        url = d.pop("url")

        starting_balance = d.pop("starting_balance")

        status = d.pop("status")

        invoice = cls(
            amount_due=amount_due,
            period_start=period_start,
            period_end=period_end,
            subtotal=subtotal,
            total=total,
            items=items,
            url=url,
            starting_balance=starting_balance,
            status=status,
        )

        invoice.additional_properties = d
        return invoice

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="CustomerDetails")


@attr.s(auto_attribs=True)
class CustomerDetails:
    """
    Attributes:
        user_id (str):
        soft_monthly_budget (Union[Unset, int]):  Default: 2500.
        hard_monthly_budget (Union[Unset, int]):  Default: 3000.
        current_balance (Union[Unset, int]):
    """

    user_id: str
    soft_monthly_budget: Union[Unset, int] = 2500
    hard_monthly_budget: Union[Unset, int] = 3000
    current_balance: Union[Unset, int] = 0
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        user_id = self.user_id
        soft_monthly_budget = self.soft_monthly_budget
        hard_monthly_budget = self.hard_monthly_budget
        current_balance = self.current_balance

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "user_id": user_id,
            }
        )
        if soft_monthly_budget is not UNSET:
            field_dict["soft_monthly_budget"] = soft_monthly_budget
        if hard_monthly_budget is not UNSET:
            field_dict["hard_monthly_budget"] = hard_monthly_budget
        if current_balance is not UNSET:
            field_dict["current_balance"] = current_balance

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        user_id = d.pop("user_id")

        soft_monthly_budget = d.pop("soft_monthly_budget", UNSET)

        hard_monthly_budget = d.pop("hard_monthly_budget", UNSET)

        current_balance = d.pop("current_balance", UNSET)

        customer_details = cls(
            user_id=user_id,
            soft_monthly_budget=soft_monthly_budget,
            hard_monthly_budget=hard_monthly_budget,
            current_balance=current_balance,
        )

        customer_details.additional_properties = d
        return customer_details

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import os
from dataclasses import dataclass, field

from fal.toolkit.file.types import FileData, FileRepository
from fal.toolkit.mainify import mainify

DEFAULT_URL_TIMEOUT = 60 * 15  # 15 minutes


@mainify
@dataclass
class GoogleStorageRepository(FileRepository):
    url_expiration: int | None = field(default_factory=lambda: DEFAULT_URL_TIMEOUT)
    bucket_name: str | None = field(default="fal_file_storage")
    gcp_account_json: str | None = None

    _storage_client = None

    def __post_init__(self):
        gcp_account_json = self.gcp_account_json
        if gcp_account_json is None:
            gcp_account_json = os.environ.get("GCLOUD_SA_JSON")
            if gcp_account_json is None:
                raise Exception("GCLOUD_SA_JSON environment secret is not set")

        import json

        gcp_account_info = json.loads(gcp_account_json)

        from google.cloud.storage import Client

        self._storage_client = Client.from_service_account_info(gcp_account_info)

    def save(self, data: FileData) -> str:
        import datetime
        from typing import cast

        from google.cloud.storage import Client

        if not isinstance(self._storage_client, Client):
            raise Exception("Google Storage client is not initialized")

        client = cast(Client, self._storage_client)
        bucket = client.get_bucket(self.bucket_name)
        gcp_blob = bucket.blob(data.file_name)
        gcp_blob.upload_from_string(data.data, content_type=data.content_type)

        if self.url_expiration is None:
            return gcp_blob.public_url

        return gcp_blob.generate_signed_url(
            version="v4",
            expiration=datetime.timedelta(minutes=self.url_expiration),
            method="GET",
        )



from __future__ import annotations


# HACK: make classes dillable https://github.com/uqfoundation/dill/issues/424
# Only works for classes for now, must be outer-most decorator in most cases
def mainify(obj):
    if hasattr(obj, "__module__") and obj.__module__.startswith("fal"):
        obj.__module__ = "__main__"

        for inner in obj.__dict__.values():
            mainify(inner)

    return obj



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import attr
from dateutil.parser import isoparse

from ..models.run_type import RunType

if TYPE_CHECKING:
    from ..models.persisted_usage_record_meta import PersistedUsageRecordMeta


T = TypeVar("T", bound="PersistedUsageRecord")


@attr.s(auto_attribs=True)
class PersistedUsageRecord:
    """
    Attributes:
        id (str):
        subscription_item_id (str):
        subscription_id (str):
        quantity (int):
        run_type (RunType): An enumeration.
        machine_type (str):
        timestamp (datetime.datetime):
        meta (PersistedUsageRecordMeta):
    """

    id: str
    subscription_item_id: str
    subscription_id: str
    quantity: int
    run_type: RunType
    machine_type: str
    timestamp: datetime.datetime
    meta: "PersistedUsageRecordMeta"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        subscription_item_id = self.subscription_item_id
        subscription_id = self.subscription_id
        quantity = self.quantity
        run_type = self.run_type.value

        machine_type = self.machine_type
        timestamp = self.timestamp.isoformat()

        meta = self.meta.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "subscription_item_id": subscription_item_id,
                "subscription_id": subscription_id,
                "quantity": quantity,
                "run_type": run_type,
                "machine_type": machine_type,
                "timestamp": timestamp,
                "meta": meta,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.persisted_usage_record_meta import PersistedUsageRecordMeta

        d = src_dict.copy()
        id = d.pop("id")

        subscription_item_id = d.pop("subscription_item_id")

        subscription_id = d.pop("subscription_id")

        quantity = d.pop("quantity")

        run_type = RunType(d.pop("run_type"))

        machine_type = d.pop("machine_type")

        timestamp = isoparse(d.pop("timestamp"))

        meta = PersistedUsageRecordMeta.from_dict(d.pop("meta"))

        persisted_usage_record = cls(
            id=id,
            subscription_item_id=subscription_item_id,
            subscription_id=subscription_id,
            quantity=quantity,
            run_type=run_type,
            machine_type=machine_type,
            timestamp=timestamp,
            meta=meta,
        )

        persisted_usage_record.additional_properties = d
        return persisted_usage_record

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar

import attr
from dateutil.parser import isoparse

T = TypeVar("T", bound="UsagePerMachineType")


@attr.s(auto_attribs=True)
class UsagePerMachineType:
    """
    Attributes:
        machine_type (str):
        day (datetime.date):
        num_workers (int):
        num_runs (int):
        run_call_seconds (int):
        run_execution_seconds (int):
        run_queue_seconds (int):
        worker_uptime_seconds (int):
        worker_idle_seconds (int):
    """

    machine_type: str
    day: datetime.date
    num_workers: int
    num_runs: int
    run_call_seconds: int
    run_execution_seconds: int
    run_queue_seconds: int
    worker_uptime_seconds: int
    worker_idle_seconds: int
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        machine_type = self.machine_type
        day = self.day.isoformat()
        num_workers = self.num_workers
        num_runs = self.num_runs
        run_call_seconds = self.run_call_seconds
        run_execution_seconds = self.run_execution_seconds
        run_queue_seconds = self.run_queue_seconds
        worker_uptime_seconds = self.worker_uptime_seconds
        worker_idle_seconds = self.worker_idle_seconds

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "machine_type": machine_type,
                "day": day,
                "num_workers": num_workers,
                "num_runs": num_runs,
                "run_call_seconds": run_call_seconds,
                "run_execution_seconds": run_execution_seconds,
                "run_queue_seconds": run_queue_seconds,
                "worker_uptime_seconds": worker_uptime_seconds,
                "worker_idle_seconds": worker_idle_seconds,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        machine_type = d.pop("machine_type")

        day = isoparse(d.pop("day")).date()

        num_workers = d.pop("num_workers")

        num_runs = d.pop("num_runs")

        run_call_seconds = d.pop("run_call_seconds")

        run_execution_seconds = d.pop("run_execution_seconds")

        run_queue_seconds = d.pop("run_queue_seconds")

        worker_uptime_seconds = d.pop("worker_uptime_seconds")

        worker_idle_seconds = d.pop("worker_idle_seconds")

        usage_per_machine_type = cls(
            machine_type=machine_type,
            day=day,
            num_workers=num_workers,
            num_runs=num_runs,
            run_call_seconds=run_call_seconds,
            run_execution_seconds=run_execution_seconds,
            run_queue_seconds=run_queue_seconds,
            worker_uptime_seconds=worker_uptime_seconds,
            worker_idle_seconds=worker_idle_seconds,
        )

        usage_per_machine_type.additional_properties = d
        return usage_per_machine_type

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="PaymentMethod")


@attr.s(auto_attribs=True)
class PaymentMethod:
    """
    Attributes:
        type (str):
        brand (str):
        last4 (str):
        exp_month (int):
        exp_year (int):
    """

    type: str
    brand: str
    last4: str
    exp_month: int
    exp_year: int
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type
        brand = self.brand
        last4 = self.last4
        exp_month = self.exp_month
        exp_year = self.exp_year

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "brand": brand,
                "last4": last4,
                "exp_month": exp_month,
                "exp_year": exp_year,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = d.pop("type")

        brand = d.pop("brand")

        last4 = d.pop("last4")

        exp_month = d.pop("exp_month")

        exp_year = d.pop("exp_year")

        payment_method = cls(
            type=type,
            brand=brand,
            last4=last4,
            exp_month=exp_month,
            exp_year=exp_year,
        )

        payment_method.additional_properties = d
        return payment_method

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import os
from base64 import b64encode
from dataclasses import dataclass

from fal.toolkit.file.types import FileData, FileRepository
from fal.toolkit.mainify import mainify


@mainify
@dataclass
class FalFileRepository(FileRepository):
    def save(self, file: FileData) -> str:
        from json import load as load_json
        from urllib.error import HTTPError
        from urllib.request import Request, urlopen
        from uuid import uuid4

        key_id = os.environ.get("FAL_KEY_ID")
        key_secret = os.environ.get("FAL_KEY_SECRET")

        boundary = f"----FormBoundary-{uuid4().hex}"
        headers = {
            "Authorization": f"Key {key_id}:{key_secret}",
            "Accept": "application/json",
            "Content-Type": f"multipart/form-data; boundary={boundary}",
        }

        grpc_host = os.environ.get("FAL_HOST", "api.alpha.fal.ai")
        rest_host = grpc_host.replace("api", "rest", 1)
        storage_url = f"https://{rest_host}/storage/upload"

        body = (
            f"--{boundary}\r\n"
            f'Content-Disposition: file; name="file"; filename="{file.file_name}"\r\n'
            f"Content-Type: {file.content_type}\r\n\r\n".encode()
            + file.data
            + f"\r\n--{boundary}--\r\n".encode()
        )

        try:
            req = Request(storage_url, data=body, headers=headers, method="POST")
            with urlopen(req) as response:
                result = load_json(response)
        except HTTPError as e:
            # TODO sometimes the error body is JSON
            # What's the appropriate way to handle it and forward the payload?
            raise Exception(f"Error uploading file. Status {e.status}: {e.reason}")

        return result["url"]


@mainify
@dataclass
class InMemoryRepository(FileRepository):
    def save(self, file: FileData) -> str:
        return f'data:{file.content_type};base64,{b64encode(file.data).decode("utf-8")}'



from __future__ import annotations

from pathlib import Path
from typing import Callable, Literal

from fal.toolkit.file.providers.fal import FalFileRepository, InMemoryRepository
from fal.toolkit.file.providers.gcp import GoogleStorageRepository
from fal.toolkit.file.types import FileData, FileRepository, RepositoryId
from fal.toolkit.mainify import mainify
from pydantic import BaseModel, Field, PrivateAttr

BuiltInRepositoryId = Literal["fal", "in_memory", "gcp_storage"]
FileRepositoryFactory = Callable[[], FileRepository]

BUILT_IN_REPOSITORIES: dict[BuiltInRepositoryId, FileRepositoryFactory] = {
    "fal": lambda: FalFileRepository(),
    "in_memory": lambda: InMemoryRepository(),
    "gcp_storage": lambda: GoogleStorageRepository(),
}


def get_builtin_repository(id: BuiltInRepositoryId) -> FileRepository:
    if id not in BUILT_IN_REPOSITORIES.keys():
        raise ValueError(f'"{id}" is not a valid built-in file repository')
    return BUILT_IN_REPOSITORIES[id]()


get_builtin_repository.__module__ = "__main__"

DEFAULT_REPOSITORY: FileRepository | BuiltInRepositoryId = "fal"


@mainify
class File(BaseModel):
    # public properties
    _file_data: FileData = PrivateAttr()
    url: str = Field(
        description="The URL where the file can be downloaded from.",
        examples=["https://url.to/generated/file/z9RV14K95DvU.png"],
    )
    content_type: str = Field(
        description="The mime type of the file.",
        examples=["image/png"],
    )
    file_name: str = Field(
        description="The name of the file. It will be auto-generated if not provided.",
        examples=["z9RV14K95DvU.png"],
    )
    file_size: int = Field(
        description="The size of the file in bytes.", examples=[4404019]
    )

    def __init__(self, **kwargs):
        if "file_data" in kwargs:
            data = kwargs.pop("file_data")
            repository = kwargs.pop("repository", None)

            repo = (
                repository
                if isinstance(repository, FileRepository)
                else get_builtin_repository(repository)
            )
            self._file_data = data

            kwargs.update(
                {
                    "url": repo.save(data),
                    "content_type": data.content_type,
                    "file_name": data.file_name,
                    "file_size": len(data.data),
                }
            )

        super().__init__(**kwargs)

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        content_type: str | None = None,
        file_name: str | None = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
    ) -> File:
        return cls(
            file_data=FileData(data, content_type, file_name),
            repository=repository,
        )

    @classmethod
    def from_path(
        cls,
        path: str | Path,
        content_type: str | None = None,
        repository: FileRepository | RepositoryId = DEFAULT_REPOSITORY,
    ) -> File:
        file_path = Path(path)
        if not file_path.exists():
            raise FileNotFoundError(f"File {file_path} does not exist")
        with open(file_path, "rb") as f:
            data = f.read()
        return File.from_bytes(
            data, content_type, file_name=file_path.name, repository=repository
        )

    def as_bytes(self) -> bytes:
        return self._file_data.data



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'openapi_fal_rest': 'openapi-fal-rest/openapi_fal_rest',
 'openapi_fal_rest.api': 'openapi-fal-rest/openapi_fal_rest/api',
 'openapi_fal_rest.api.billing': 'openapi-fal-rest/openapi_fal_rest/api/billing',
 'openapi_fal_rest.api.files': 'openapi-fal-rest/openapi_fal_rest/api/files',
 'openapi_fal_rest.api.health': 'openapi-fal-rest/openapi_fal_rest/api/health',
 'openapi_fal_rest.api.keys': 'openapi-fal-rest/openapi_fal_rest/api/keys',
 'openapi_fal_rest.api.logs': 'openapi-fal-rest/openapi_fal_rest/api/logs',
 'openapi_fal_rest.api.storage': 'openapi-fal-rest/openapi_fal_rest/api/storage',
 'openapi_fal_rest.api.usage': 'openapi-fal-rest/openapi_fal_rest/api/usage',
 'openapi_fal_rest.models': 'openapi-fal-rest/openapi_fal_rest/models'}

packages = \
['fal',
 'fal.auth',
 'fal.console',
 'fal.exceptions',
 'fal.logging',
 'fal.toolkit',
 'fal.toolkit.file',
 'fal.toolkit.file.providers',
 'fal.toolkit.image',
 'openapi_fal_rest',
 'openapi_fal_rest.api',
 'openapi_fal_rest.api.billing',
 'openapi_fal_rest.api.files',
 'openapi_fal_rest.api.health',
 'openapi_fal_rest.api.keys',
 'openapi_fal_rest.api.logs',
 'openapi_fal_rest.api.storage',
 'openapi_fal_rest.api.usage',
 'openapi_fal_rest.models']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.3.0',
 'auth0-python>=4.1.0,<5.0.0',
 'click>=8.1.3,<9.0.0',
 'colorama>=0.4.6,<0.5.0',
 'datadog-api-client==2.12.0',
 'dill==0.3.7',
 'grpc-interceptor>=0.15.0,<0.16.0',
 'grpcio>=1.50.0,<2.0.0',
 'httpx>=0.15.4,<0.25.0',
 'isolate-proto>=0.0.44,<0.0.45',
 'isolate[build]>=0.12.2,<1.0',
 'opentelemetry-api>=1.15.0,<2.0.0',
 'opentelemetry-sdk>=1.15.0,<2.0.0',
 'packaging>=21.3',
 'pathspec>=0.11.1,<0.12.0',
 'portalocker>=2.7.0,<3.0.0',
 'pydantic<2.0',
 'python-dateutil>=2.8.0,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=13.3.2,<14.0.0',
 'structlog>=22.3.0,<23.0.0',
 'types-python-dateutil>=2.8.0,<3.0.0',
 'typing-extensions>=4.7.1,<5.0.0']

extras_require = \
{':python_version < "3.10"': ['importlib-metadata>=4.4']}

entry_points = \
{'console_scripts': ['fal = fal.cli:cli', 'fal-serverless = fal.cli:cli']}

setup_kwargs = {
    'name': 'fal',
    'version': '0.10.7',
    'description': 'fal is an easy-to-use Serverless Python Framework',
    'long_description': "# fal\n\nfal is a serverless Python runtime that lets you run and scale code in the cloud with no infra management.\n\nWith fal, you can build pipelines, serve ML models and scale them up to many users. You scale down to 0 when you don't use any resources.\n\nCheck out to the [docs](https://serverless.fal.ai/docs) for more details.\n",
    'author': 'Features & Labels',
    'author_email': 'hello@fal.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="FileSpec")


@attr.s(auto_attribs=True)
class FileSpec:
    """
    Attributes:
        path (str):
        name (str):
        created_time (float):
        updated_time (float):
        is_file (bool):
        size (int):
        checksum_sha256 (Union[Unset, str]):
        checksum_md5 (Union[Unset, str]):
    """

    path: str
    name: str
    created_time: float
    updated_time: float
    is_file: bool
    size: int
    checksum_sha256: Union[Unset, str] = UNSET
    checksum_md5: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        path = self.path
        name = self.name
        created_time = self.created_time
        updated_time = self.updated_time
        is_file = self.is_file
        size = self.size
        checksum_sha256 = self.checksum_sha256
        checksum_md5 = self.checksum_md5

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "path": path,
                "name": name,
                "created_time": created_time,
                "updated_time": updated_time,
                "is_file": is_file,
                "size": size,
            }
        )
        if checksum_sha256 is not UNSET:
            field_dict["checksum_sha256"] = checksum_sha256
        if checksum_md5 is not UNSET:
            field_dict["checksum_md5"] = checksum_md5

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        path = d.pop("path")

        name = d.pop("name")

        created_time = d.pop("created_time")

        updated_time = d.pop("updated_time")

        is_file = d.pop("is_file")

        size = d.pop("size")

        checksum_sha256 = d.pop("checksum_sha256", UNSET)

        checksum_md5 = d.pop("checksum_md5", UNSET)

        file_spec = cls(
            path=path,
            name=name,
            created_time=created_time,
            updated_time=updated_time,
            is_file=is_file,
            size=size,
            checksum_sha256=checksum_sha256,
            checksum_md5=checksum_md5,
        )

        file_spec.additional_properties = d
        return file_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
